import argparse
import json
from pprint import pprint
import sys


def pprint_json(args: argparse.Namespace) -> int:
    """
    Prints the contents of a JSON file using pprint.
    """
    try:
        with open(args.file_path, "r") as file:
            data = json.load(file)
        pprint(data)

    except FileNotFoundError:
        print(f"File not found: {args.file_path}")
        return 1
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON: {e}")
        return 1

    return 0


def execute(args: argparse.Namespace) -> int:
    result = args.func(args)
    if not isinstance(result, int):
        raise TypeError(
            f"Each con-duct subcommand must return an int returncode, got {type(result)}"
        )
    return result


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="con-duct",
        description="A command-line tool for managing various tasks.",
        usage="con-duct <command> [options]",
    )
    subparsers = parser.add_subparsers(dest="command", help="Available subcommands")

    # Subcommand: pp
    parser_pp = subparsers.add_parser("pp", help="Pretty print a JSON log")
    parser_pp.add_argument("file_path", help="JSON file to pretty print")
    parser_pp.set_defaults(func=pprint_json)

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
    else:
        sys.exit(execute(args))



#!/usr/bin/env python3
from __future__ import annotations
import argparse
from collections import Counter
from collections.abc import Iterable, Iterator
from dataclasses import asdict, dataclass, field
from datetime import datetime
from enum import Enum
import json
import logging
import math
import os
import re
import shutil
import socket
import string
import subprocess
import sys
import textwrap
import threading
import time
from typing import IO, Any, Optional, TextIO

__version__ = "0.6.0"
__schema_version__ = "0.1.1"


lgr = logging.getLogger("con-duct")
DEFAULT_LOG_LEVEL = os.environ.get("DUCT_LOG_LEVEL", "INFO").upper()

ENV_PREFIXES = ("PBS_", "SLURM_", "OSG")
SUFFIXES = {
    "stdout": "stdout",
    "stderr": "stderr",
    "usage": "usage.json",
    "info": "info.json",
}
EXECUTION_SUMMARY_FORMAT = (
    "Summary:\n"
    "Exit Code: {exit_code!E}\n"
    "Command: {command}\n"
    "Log files location: {logs_prefix}\n"
    "Wall Clock Time: {wall_clock_time:.3f} sec\n"
    "Memory Peak Usage (RSS): {peak_rss!S}\n"
    "Memory Average Usage (RSS): {average_rss!S}\n"
    "Virtual Memory Peak Usage (VSZ): {peak_vsz!S}\n"
    "Virtual Memory Average Usage (VSZ): {average_vsz!S}\n"
    "Memory Peak Percentage: {peak_pmem!N}%\n"
    "Memory Average Percentage: {average_pmem!N}%\n"
    "CPU Peak Usage: {peak_pcpu!N}%\n"
    "Average CPU Usage: {average_pcpu!N}%\n"
    "Samples Collected: {num_samples!X}\n"
    "Reports Written: {num_reports!X}\n"
)


ABOUT_DUCT = """
duct is a lightweight wrapper that collects execution data for an arbitrary
command.  Execution data includes execution time, system information, and
resource usage statistics of the command and all its child processes. It is
intended to simplify the problem of recording the resources necessary to
execute a command, particularly in an HPC environment.

Resource usage is determined by polling (at a sample-interval).
During execution, duct produces a JSON lines (see https://jsonlines.org) file
with one data point recorded for each report (at a report-interval).

environment variables:
  Many duct options can be configured by environment variables (which are
  overridden by command line options).

  DUCT_LOG_LEVEL: see --log-level
  DUCT_OUTPUT_PREFIX: see --output-prefix
  DUCT_SUMMARY_FORMAT: see --summary-format
  DUCT_SAMPLE_INTERVAL: see --sample-interval
  DUCT_REPORT_INTERVAL: see --report-interval
  DUCT_CAPTURE_OUTPUTS: see --capture-outputs
"""


class CustomHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _fill_text(self, text: str, width: int, _indent: str) -> str:
        # Override _fill_text to respect the newlines and indentation in descriptions
        return "\n".join([textwrap.fill(line, width) for line in text.splitlines()])


def assert_num(*values: Any) -> None:
    for value in values:
        assert isinstance(value, (float, int))


class Outputs(str, Enum):
    ALL = "all"
    NONE = "none"
    STDOUT = "stdout"
    STDERR = "stderr"

    def __str__(self) -> str:
        return self.value

    def has_stdout(self) -> bool:
        return self is Outputs.ALL or self is Outputs.STDOUT

    def has_stderr(self) -> bool:
        return self is Outputs.ALL or self is Outputs.STDERR


class RecordTypes(str, Enum):
    ALL = "all"
    SYSTEM_SUMMARY = "system-summary"
    PROCESSES_SAMPLES = "processes-samples"

    def __str__(self) -> str:
        return self.value

    def has_system_summary(self) -> bool:
        return self is RecordTypes.ALL or self is RecordTypes.SYSTEM_SUMMARY

    def has_processes_samples(self) -> bool:
        return self is RecordTypes.ALL or self is RecordTypes.PROCESSES_SAMPLES


@dataclass
class SystemInfo:
    cpu_total: int
    memory_total: int
    hostname: str | None
    uid: int
    user: str | None


@dataclass
class ProcessStats:
    pcpu: float  # %CPU
    pmem: float  # %MEM
    rss: int  # Memory Resident Set Size in Bytes
    vsz: int  # Virtual Memory size in Bytes
    timestamp: str
    etime: str
    stat: Counter
    cmd: str

    def aggregate(self, other: ProcessStats) -> ProcessStats:
        cmd = self.cmd
        if self.cmd != other.cmd:
            lgr.debug(
                f"cmd has changed. Previous measurement was {self.cmd}, now {other.cmd}."
            )
            # Brackets indicate that the kernel has substituted an abbreviation.
            surrounded_by_brackets = r"^\[.+\]"
            if re.search(surrounded_by_brackets, self.cmd):
                lgr.debug(f"using {other.cmd}.")
                cmd = other.cmd
            lgr.debug(f"using {self.cmd}.")

        new_counter: Counter = Counter()
        new_counter.update(self.stat)
        new_counter.update(other.stat)
        return ProcessStats(
            pcpu=max(self.pcpu, other.pcpu),
            pmem=max(self.pmem, other.pmem),
            rss=max(self.rss, other.rss),
            vsz=max(self.vsz, other.vsz),
            timestamp=max(self.timestamp, other.timestamp),
            etime=other.etime,  # For the aggregate always take the latest
            stat=new_counter,
            cmd=cmd,
        )

    def for_json(self) -> dict:
        ret = asdict(self)
        ret["stat"] = dict(self.stat)
        return ret

    def __post_init__(self) -> None:
        self._validate()

    def _validate(self) -> None:
        assert_num(self.pcpu, self.pmem, self.rss, self.vsz)


@dataclass
class LogPaths:
    stdout: str
    stderr: str
    usage: str
    info: str
    prefix: str

    def __iter__(self) -> Iterator[tuple[str, str]]:
        for name, path in asdict(self).items():
            if name != "prefix":
                yield name, path

    @classmethod
    def create(cls, output_prefix: str, pid: None | int = None) -> LogPaths:
        datetime_filesafe = datetime.now().strftime("%Y.%m.%dT%H.%M.%S")
        formatted_prefix = output_prefix.format(
            pid=pid, datetime_filesafe=datetime_filesafe
        )
        return cls(
            stdout=f"{formatted_prefix}{SUFFIXES['stdout']}",
            stderr=f"{formatted_prefix}{SUFFIXES['stderr']}",
            usage=f"{formatted_prefix}{SUFFIXES['usage']}",
            info=f"{formatted_prefix}{SUFFIXES['info']}",
            prefix=formatted_prefix,
        )

    def prepare_paths(self, clobber: bool, capture_outputs: Outputs) -> None:
        conflicts = [path for _name, path in self if os.path.exists(path)]
        if conflicts and not clobber:
            raise FileExistsError(
                "Conflicting files:\n"
                + "\n".join(f"- {path}" for path in conflicts)
                + "\nUse --clobber to overrwrite conflicting files."
            )

        if self.prefix.endswith(os.sep):  # If it ends in "/" (for linux) treat as a dir
            os.makedirs(self.prefix, exist_ok=True)
        else:
            # Path does not end with a separator, treat the last part as a filename
            directory = os.path.dirname(self.prefix)
            if directory:
                os.makedirs(directory, exist_ok=True)
        for name, path in self:
            if name == SUFFIXES["stdout"] and not capture_outputs.has_stdout():
                continue
            elif name == SUFFIXES["stderr"] and not capture_outputs.has_stderr():
                continue
            # usage and info should always be created
            open(path, "w").close()


@dataclass
class Averages:
    rss: Optional[float] = None
    vsz: Optional[float] = None
    pmem: Optional[float] = None
    pcpu: Optional[float] = None
    num_samples: int = 0

    def update(self: Averages, other: Sample) -> None:
        assert_num(other.total_rss, other.total_vsz, other.total_pmem, other.total_pcpu)
        if not self.num_samples:
            self.num_samples += 1
            self.rss = other.total_rss
            self.vsz = other.total_vsz
            self.pmem = other.total_pmem
            self.pcpu = other.total_pcpu
        else:
            assert self.rss is not None
            assert self.vsz is not None
            assert self.pmem is not None
            assert self.pcpu is not None
            assert other.total_rss is not None
            assert other.total_vsz is not None
            assert other.total_pmem is not None
            assert other.total_pcpu is not None
            self.num_samples += 1
            self.rss += (other.total_rss - self.rss) / self.num_samples
            self.vsz += (other.total_vsz - self.vsz) / self.num_samples
            self.pmem += (other.total_pmem - self.pmem) / self.num_samples
            self.pcpu += (other.total_pcpu - self.pcpu) / self.num_samples

    @classmethod
    def from_sample(cls, sample: Sample) -> Averages:
        assert_num(
            sample.total_rss, sample.total_vsz, sample.total_pmem, sample.total_pcpu
        )
        return cls(
            rss=sample.total_rss,
            vsz=sample.total_vsz,
            pmem=sample.total_pmem,
            pcpu=sample.total_pcpu,
            num_samples=1,
        )


@dataclass
class Sample:
    stats: dict[int, ProcessStats] = field(default_factory=dict)
    averages: Averages = field(default_factory=Averages)
    total_rss: Optional[int] = None
    total_vsz: Optional[int] = None
    total_pmem: Optional[float] = None
    total_pcpu: Optional[float] = None
    timestamp: str = ""  # TS of last sample collected

    def add_pid(self, pid: int, stats: ProcessStats) -> None:
        # We do not calculate averages when we add a pid because we require all pids first
        assert (
            self.stats.get(pid) is None
        )  # add_pid should only be called when pid not in Sample
        self.total_rss = (self.total_rss or 0) + stats.rss
        self.total_vsz = (self.total_vsz or 0) + stats.vsz
        self.total_pmem = (self.total_pmem or 0.0) + stats.pmem
        self.total_pcpu = (self.total_pcpu or 0.0) + stats.pcpu
        self.stats[pid] = stats
        self.timestamp = max(self.timestamp, stats.timestamp)

    def aggregate(self: Sample, other: Sample) -> Sample:
        output = Sample()
        for pid in self.stats.keys() | other.stats.keys():
            if (mine := self.stats.get(pid)) is not None:
                if (theirs := other.stats.get(pid)) is not None:
                    output.add_pid(pid, mine.aggregate(theirs))
                else:
                    output.add_pid(pid, mine)
            else:
                output.add_pid(pid, other.stats[pid])
        assert other.total_pmem is not None
        assert other.total_pcpu is not None
        assert other.total_rss is not None
        assert other.total_vsz is not None
        output.total_pmem = max(self.total_pmem or 0.0, other.total_pmem)
        output.total_pcpu = max(self.total_pcpu or 0.0, other.total_pcpu)
        output.total_rss = max(self.total_rss or 0, other.total_rss)
        output.total_vsz = max(self.total_vsz or 0, other.total_vsz)
        output.averages = self.averages
        output.averages.update(other)
        return output

    def for_json(self) -> dict[str, Any]:
        d = {
            "timestamp": self.timestamp,
            "num_samples": self.averages.num_samples,
            "processes": {
                str(pid): stats.for_json() for pid, stats in self.stats.items()
            },
            "totals": {  # total of all processes during this sample
                "pmem": self.total_pmem,
                "pcpu": self.total_pcpu,
                "rss": self.total_rss,
                "vsz": self.total_vsz,
            },
            "averages": asdict(self.averages) if self.averages.num_samples >= 1 else {},
        }
        return d


class Report:
    """Top level report"""

    def __init__(
        self,
        command: str,
        arguments: list[str],
        log_paths: LogPaths,
        summary_format: str,
        colors: bool = False,
        clobber: bool = False,
        process: subprocess.Popen | None = None,
    ) -> None:
        self._command = command
        self.arguments = arguments
        self.log_paths = log_paths
        self.summary_format: str = summary_format
        self.clobber = clobber
        self.colors = colors
        # Defaults to be set later
        self.start_time: float | None = None
        self.process = process
        self.session_id: int | None = None
        self.gpus: list[dict[str, str]] | None = None
        self.env: dict[str, str] | None = None
        self.number = 1
        self.system_info: SystemInfo | None = None
        self.full_run_stats = Sample()
        self.current_sample: Optional[Sample] = None
        self.end_time: float | None = None
        self.run_time_seconds: str | None = None

    @property
    def command(self) -> str:
        return " ".join([self._command] + self.arguments)

    @property
    def elapsed_time(self) -> float:
        assert self.start_time is not None
        return time.time() - self.start_time

    @property
    def wall_clock_time(self) -> Optional[float]:
        if self.start_time is None:
            return math.nan
        if self.end_time is None:
            # if no end_time -- must be still ongoing
            # Cannot happen ATM but could in "library mode" later
            return time.time() - self.start_time
        # we reached the end
        return self.end_time - self.start_time

    def collect_environment(self) -> None:
        self.env = {k: v for k, v in os.environ.items() if k.startswith(ENV_PREFIXES)}

    def get_system_info(self) -> None:
        """Gathers system information related to CPU, GPU, memory, and environment variables."""
        self.system_info = SystemInfo(
            cpu_total=os.sysconf("SC_NPROCESSORS_CONF"),
            memory_total=os.sysconf("SC_PAGESIZE") * os.sysconf("SC_PHYS_PAGES"),
            hostname=socket.gethostname(),
            uid=os.getuid(),
            user=os.environ.get("USER"),
        )
        # GPU information
        if shutil.which("nvidia-smi") is not None:
            lgr.debug("Checking NVIDIA GPU using nvidia-smi")
            try:
                out = subprocess.check_output(
                    [
                        "nvidia-smi",
                        "--query-gpu=index,name,pci.bus_id,driver_version,memory.total,compute_mode",
                        "--format=csv",
                    ]
                )
            except subprocess.CalledProcessError as e:
                lgr.warning("Error collecting gpu information: %s", str(e))
                self.gpus = None
                return
            try:
                decoded = out.decode("utf-8")
                lines = decoded.strip().split("\n")
                _ = lines.pop(0)  # header
                self.gpus = []
                for line in lines:
                    cols = line.split(", ")
                    self.gpus.append(
                        {
                            "index": cols[0],
                            "name": cols[1],
                            "bus_id": cols[2],
                            "driver_version": cols[3],
                            "memory.total": cols[4],
                            "compute_mode": cols[5],
                        }
                    )
            except Exception as e:
                lgr.warning("Error parsing gpu information: %s", str(e))
                self.gpus = None

    def collect_sample(self) -> Optional[Sample]:
        assert self.session_id is not None
        sample = Sample()
        try:
            output = subprocess.check_output(
                [
                    "ps",
                    "-s",
                    str(self.session_id),
                    "-o",
                    "pid,pcpu,pmem,rss,vsz,etime,stat,cmd",
                ],
                text=True,
            )
            for line in output.splitlines()[1:]:
                if line:
                    pid, pcpu, pmem, rss_kib, vsz_kib, etime, stat, cmd = line.split(
                        maxsplit=7,
                    )
                    sample.add_pid(
                        int(pid),
                        ProcessStats(
                            pcpu=float(pcpu),
                            pmem=float(pmem),
                            rss=int(rss_kib) * 1024,
                            vsz=int(vsz_kib) * 1024,
                            timestamp=datetime.now().astimezone().isoformat(),
                            etime=etime,
                            stat=Counter([stat]),
                            cmd=cmd,
                        ),
                    )
        except subprocess.CalledProcessError as exc:  # when session_id has no processes
            lgr.debug("Error collecting sample: %s", str(exc))
            return None

        sample.averages = Averages.from_sample(sample)
        return sample

    def update_from_sample(self, sample: Sample) -> None:
        self.full_run_stats = self.full_run_stats.aggregate(sample)
        if self.current_sample is None:
            self.current_sample = Sample().aggregate(sample)
        else:
            assert self.current_sample.averages is not None
            self.current_sample = self.current_sample.aggregate(sample)
        assert self.current_sample is not None

    def write_subreport(self) -> None:
        assert self.current_sample is not None
        with open(self.log_paths.usage, "a") as resource_statistics_log:
            resource_statistics_log.write(
                json.dumps(self.current_sample.for_json()) + "\n"
            )

    @property
    def execution_summary(self) -> dict[str, Any]:
        # killed by a signal
        # https://pubs.opengroup.org/onlinepubs/9799919799/utilities/V3_chap02.html#tag_19_08_02
        if self.process and self.process.returncode < 0:
            self.process.returncode = 128 + abs(self.process.returncode)
        # prepare the base, but enrich if we did get process running
        return {
            "exit_code": self.process.returncode if self.process else None,
            "command": self.command,
            "logs_prefix": self.log_paths.prefix if self.log_paths else "",
            "wall_clock_time": self.wall_clock_time,
            "peak_rss": self.full_run_stats.total_rss,
            "average_rss": self.full_run_stats.averages.rss,
            "peak_vsz": self.full_run_stats.total_vsz,
            "average_vsz": self.full_run_stats.averages.vsz,
            "peak_pmem": self.full_run_stats.total_pmem,
            "average_pmem": self.full_run_stats.averages.pmem,
            "peak_pcpu": self.full_run_stats.total_pcpu,
            "average_pcpu": self.full_run_stats.averages.pcpu,
            "num_samples": self.full_run_stats.averages.num_samples,
            "num_reports": self.number,
        }

    def dump_json(self) -> str:
        return json.dumps(
            {
                "command": self.command,
                "system": (
                    None if self.system_info is None else asdict(self.system_info)
                ),
                "env": self.env,
                "gpu": self.gpus,
                "duct_version": __version__,
                "schema_version": __schema_version__,
                "execution_summary": self.execution_summary,
                "output_paths": asdict(self.log_paths),
            }
        )

    @property
    def execution_summary_formatted(self) -> str:
        formatter = SummaryFormatter(enable_colors=self.colors)
        return formatter.format(self.summary_format, **self.execution_summary)


class SummaryFormatter(string.Formatter):
    OK = "OK"
    NOK = "X"
    NONE = "-"
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)
    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    FILESIZE_SUFFIXES = (" kB", " MB", " GB", " TB", " PB", " EB", " ZB", " YB")

    def __init__(self, enable_colors: bool = False) -> None:
        self.enable_colors = enable_colors

    def naturalsize(
        self,
        value: float | str,
        format: str = "%.1f",  # noqa: A002
    ) -> str:
        """Format a number of bytes like a human readable decimal filesize (e.g. 10 kB).

        Examples:
            ```pycon
            >>> formatter = SummaryFormatter()
            >>> formatter.naturalsize(3000000)
            '3.0 MB'
            >>> formatter.naturalsize(3000, "%.3f")
            '2.930 kB'
            >>> formatter.naturalsize(10**28)
            '10000.0 YB'
            ```

        Args:
            value (int, float, str): Integer to convert.
            format (str): Custom formatter.

        Returns:
            str: Human readable representation of a filesize.
        """
        base = 1000
        bytes_ = float(value)
        abs_bytes = abs(bytes_)

        if abs_bytes == 1:
            return "%d Byte" % bytes_

        if abs_bytes < base:
            return "%d Bytes" % bytes_

        for i, _s in enumerate(self.FILESIZE_SUFFIXES):
            unit = base ** (i + 2)

            if abs_bytes < unit:
                break

        ret: str = format % (base * bytes_ / unit) + _s
        return ret

    def color_word(self, s: str, color: int) -> str:
        """Color `s` with `color`.

        Parameters
        ----------
        s : string
        color : int
            Code for color. If the value evaluates to false, the string will not be
            colored.
        enable_colors: boolean, optional

        Returns
        -------
        str
        """
        if color and self.enable_colors:
            return "%s%s%s" % (self.COLOR_SEQ % color, s, self.RESET_SEQ)
        return s

    def convert_field(self, value: str | None, conversion: str | None) -> Any:
        if conversion == "S":  # Human size
            if value is not None:
                return self.color_word(self.naturalsize(value), self.GREEN)
            else:
                return self.color_word(self.NONE, self.RED)
        elif conversion == "E":  # colored non-zero is bad
            return self.color_word(
                value if value is not None else self.NONE,
                self.RED if value or value is None else self.GREEN,
            )
        elif conversion == "X":  # colored truthy
            col = self.GREEN if value else self.RED
            return self.color_word(value if value is not None else self.NONE, col)
        elif conversion == "N":  # colored Red - if None
            if value is None:
                return self.color_word(self.NONE, self.RED)
            else:
                return self.color_word(value, self.GREEN)

        return super().convert_field(value, conversion)

    def format_field(self, value: Any, format_spec: str) -> Any:
        # TODO: move all the "coloring" into formatting, so we could correctly indent
        # given the format and only then color it up
        # print "> %r, %r" % (value, format_spec)
        if value is None:
            # TODO: could still use our formatter and make it red or smth like that
            return self.NONE
        try:
            return super().format_field(value, format_spec)
        except ValueError:
            lgr.warning(
                f"Value: {value} is invalid for format spec {format_spec}, falling back to `str`"
            )
            return str(value)


@dataclass
class Arguments:
    command: str
    command_args: list[str]
    output_prefix: str
    sample_interval: float
    report_interval: float
    clobber: bool
    capture_outputs: Outputs
    outputs: Outputs
    record_types: RecordTypes
    summary_format: str
    colors: bool
    log_level: str
    quiet: bool

    def __post_init__(self) -> None:
        if self.report_interval < self.sample_interval:
            raise argparse.ArgumentError(
                None,
                "--report-interval must be greater than or equal to --sample-interval.",
            )

    @classmethod
    def from_argv(
        cls, cli_args: Optional[list[str]] = None, **cli_kwargs: Any
    ) -> Arguments:
        parser = argparse.ArgumentParser(
            allow_abbrev=False,
            description=ABOUT_DUCT,
            formatter_class=CustomHelpFormatter,
        )
        parser.add_argument(
            "command",
            metavar="command [command_args ...]",
            help="The command to execute, along with its arguments.",
        )
        parser.add_argument(
            "--version", action="version", version=f"%(prog)s {__version__}"
        )
        parser.add_argument(
            "command_args", nargs=argparse.REMAINDER, help="Arguments for the command."
        )
        parser.add_argument(
            "-p",
            "--output-prefix",
            type=str,
            default=os.getenv(
                "DUCT_OUTPUT_PREFIX", ".duct/logs/{datetime_filesafe}-{pid}_"
            ),
            help="File string format to be used as a prefix for the files -- the captured "
            "stdout and stderr and the resource usage logs. The understood variables are "
            "{datetime}, {datetime_filesafe}, and {pid}. "
            "Leading directories will be created if they do not exist. "
            "You can also provide value via DUCT_OUTPUT_PREFIX env variable. ",
        )
        parser.add_argument(
            "--summary-format",
            type=str,
            default=os.getenv("DUCT_SUMMARY_FORMAT", EXECUTION_SUMMARY_FORMAT),
            help="Output template to use when printing the summary following execution. "
            "Accepts custom conversion flags: "
            "!S: Converts filesizes to human readable units, green if measured, red if None. "
            "!E: Colors exit code, green if falsey, red if truthy, and red if None. "
            "!X: Colors green if truthy, red if falsey. "
            "!N: Colors green if not None, red if None",
        )
        parser.add_argument(
            "--colors",
            action="store_true",
            default=os.getenv("DUCT_COLORS", False),
            help="Use colors in duct output.",
        )
        parser.add_argument(
            "--clobber",
            action="store_true",
            help="Replace log files if they already exist.",
        )
        parser.add_argument(
            "-l",
            "--log_level",
            default=DEFAULT_LOG_LEVEL,
            choices=("NONE", "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
            help="Level of log output to stderr, use NONE to entirely disable.",
        )
        parser.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            help="[deprecated, use log level NONE] Disable duct logging output (to stderr)",
        )
        parser.add_argument(
            "--sample-interval",
            "--s-i",
            type=float,
            default=float(os.getenv("DUCT_SAMPLE_INTERVAL", "1.0")),
            help="Interval in seconds between status checks of the running process. "
            "Sample interval must be less than or equal to report interval, and it achieves the "
            "best results when sample is significantly less than the runtime of the process.",
        )
        parser.add_argument(
            "--report-interval",
            "--r-i",
            type=float,
            default=float(os.getenv("DUCT_REPORT_INTERVAL", "60.0")),
            help="Interval in seconds at which to report aggregated data.",
        )
        parser.add_argument(
            "-c",
            "--capture-outputs",
            default=os.getenv("DUCT_CAPTURE_OUTPUTS", "all"),
            choices=list(Outputs),
            type=Outputs,
            help="Record stdout, stderr, all, or none to log files. "
            "You can also provide value via DUCT_CAPTURE_OUTPUTS env variable.",
        )
        parser.add_argument(
            "-o",
            "--outputs",
            default="all",
            choices=list(Outputs),
            type=Outputs,
            help="Print stdout, stderr, all, or none to stdout/stderr respectively.",
        )
        parser.add_argument(
            "-t",
            "--record-types",
            default="all",
            choices=list(RecordTypes),
            type=RecordTypes,
            help="Record system-summary, processes-samples, or all",
        )
        args = parser.parse_args(
            args=cli_args,
            namespace=cli_kwargs and argparse.Namespace(**cli_kwargs) or None,
        )
        return cls(
            command=args.command,
            command_args=args.command_args,
            output_prefix=args.output_prefix,
            sample_interval=args.sample_interval,
            report_interval=args.report_interval,
            capture_outputs=args.capture_outputs,
            outputs=args.outputs,
            record_types=args.record_types,
            summary_format=args.summary_format,
            clobber=args.clobber,
            colors=args.colors,
            log_level=args.log_level,
            quiet=args.quiet,
        )


def monitor_process(
    report: Report,
    process: subprocess.Popen,
    report_interval: float,
    sample_interval: float,
    stop_event: threading.Event,
) -> None:
    lgr.debug(
        "Starting monitoring of the process %s on sample interval %f for report interval %f",
        process,
        sample_interval,
        report_interval,
    )
    while True:
        if process.poll() is not None:
            lgr.debug(
                "Breaking out of the monitor since the passthrough command has finished"
            )
            break
        sample = report.collect_sample()
        # Report averages should be updated prior to sample aggregation
        if (
            sample is None
        ):  # passthrough has probably finished before sample could be collected
            if process.poll() is not None:
                lgr.debug(
                    "Breaking out of the monitor since the passthrough command has finished "
                    "before we could collect sample"
                )
                break
            # process is still running, but we could not collect sample
            continue
        report.update_from_sample(sample)
        if (
            report.start_time
            and report.elapsed_time >= (report.number - 1) * report_interval
        ):
            report.write_subreport()
            report.current_sample = None
            report.number += 1
        if stop_event.wait(timeout=sample_interval):
            lgr.debug("Breaking out because stop event was set")
            break


class TailPipe:
    """TailPipe simultaneously streams to an output stream (stdout or stderr) and a specified file."""

    TAIL_CYCLE_TIME = 0.01

    def __init__(self, file_path: str, buffer: IO[bytes]) -> None:
        self.file_path = file_path
        self.buffer = buffer
        self.stop_event: threading.Event | None = None
        self.infile: IO[bytes] | None = None
        self.thread: threading.Thread | None = None

    def start(self) -> None:
        self.stop_event = threading.Event()
        self.infile = open(self.file_path, "rb")
        self.thread = threading.Thread(target=self._tail, daemon=True)
        self.thread.start()

    def fileno(self) -> int:
        assert self.infile is not None
        return self.infile.fileno()

    def _catch_up(self) -> None:
        assert self.infile is not None
        data = self.infile.read()
        if data:
            self.buffer.write(data)
            self.buffer.flush()

    def _tail(self) -> None:
        assert self.stop_event is not None
        try:
            while not self.stop_event.is_set():
                self._catch_up()
                time.sleep(TailPipe.TAIL_CYCLE_TIME)
            # After stop event, collect and passthrough data one last time
            self._catch_up()
        except Exception:
            raise
        finally:
            self.buffer.flush()

    def close(self) -> None:
        assert self.stop_event is not None
        assert self.thread is not None
        assert self.infile is not None
        self.stop_event.set()
        self.thread.join()
        self.infile.close()


def prepare_outputs(
    capture_outputs: Outputs,
    outputs: Outputs,
    log_paths: LogPaths,
) -> tuple[TextIO | TailPipe | int | None, TextIO | TailPipe | int | None]:
    stdout: TextIO | TailPipe | int | None
    stderr: TextIO | TailPipe | int | None

    if capture_outputs.has_stdout():
        if outputs.has_stdout():
            stdout = TailPipe(log_paths.stdout, buffer=sys.stdout.buffer)
            stdout.start()
        else:
            stdout = open(log_paths.stdout, "w")
    elif outputs.has_stdout():
        stdout = None
    else:
        stdout = subprocess.DEVNULL

    if capture_outputs.has_stderr():
        if outputs.has_stderr():
            stderr = TailPipe(log_paths.stderr, buffer=sys.stderr.buffer)
            stderr.start()
        else:
            stderr = open(log_paths.stderr, "w")
    elif outputs.has_stderr():
        stderr = None
    else:
        stderr = subprocess.DEVNULL
    return stdout, stderr


def safe_close_files(file_list: Iterable[Any]) -> None:
    for f in file_list:
        try:
            f.close()
        except Exception:
            pass


def main() -> None:
    logging.basicConfig(
        format="%(asctime)s [%(levelname)-8s] %(name)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=getattr(logging, DEFAULT_LOG_LEVEL),
    )
    args = Arguments.from_argv()
    sys.exit(execute(args))


def execute(args: Arguments) -> int:
    """A wrapper to execute a command, monitor and log the process details.

    Returns exit code of the executed process.
    """
    if args.log_level == "NONE" or args.quiet:
        lgr.disabled = True
    else:
        lgr.setLevel(args.log_level)
    log_paths = LogPaths.create(args.output_prefix, pid=os.getpid())
    log_paths.prepare_paths(args.clobber, args.capture_outputs)
    stdout, stderr = prepare_outputs(args.capture_outputs, args.outputs, log_paths)
    stdout_file: TextIO | IO[bytes] | int | None
    if isinstance(stdout, TailPipe):
        stdout_file = open(stdout.file_path, "wb")
    else:
        stdout_file = stdout
    stderr_file: TextIO | IO[bytes] | int | None
    if isinstance(stderr, TailPipe):
        stderr_file = open(stderr.file_path, "wb")
    else:
        stderr_file = stderr

    full_command = " ".join([str(args.command)] + args.command_args)
    files_to_close = [stdout_file, stdout, stderr_file, stderr]

    report = Report(
        args.command,
        args.command_args,
        log_paths,
        args.summary_format,
        args.colors,
        args.clobber,
    )

    report.start_time = time.time()
    try:
        report.process = process = subprocess.Popen(
            [str(args.command)] + args.command_args,
            stdout=stdout_file,
            stderr=stderr_file,
            start_new_session=True,
        )
    except FileNotFoundError:
        # We failed to execute due to file not found in PATH
        # We should remove log etc files since they are 0-sized
        # degenerates etc
        safe_close_files(files_to_close)
        for _, file_path in log_paths:
            if os.path.exists(file_path):
                assert os.stat(file_path).st_size == 0
                os.remove(file_path)
        # mimicking behavior of bash and zsh.
        print(f"{args.command}: command not found", file=sys.stderr)
        return 127  # seems what zsh and bash return then

    lgr.info("duct is executing %r...", full_command)
    lgr.info("Log files will be written to %s", log_paths.prefix)
    try:
        report.session_id = os.getsid(process.pid)  # Get session ID of the new process
    except ProcessLookupError:  # process has already finished
        # TODO: log this at least.
        pass
    stop_event = threading.Event()
    if args.record_types.has_processes_samples():
        monitoring_args = [
            report,
            process,
            args.report_interval,
            args.sample_interval,
            stop_event,
        ]
        monitoring_thread = threading.Thread(
            target=monitor_process, args=monitoring_args
        )
        monitoring_thread.start()
    else:
        monitoring_thread = None

    if args.record_types.has_system_summary():
        env_thread = threading.Thread(target=report.collect_environment)
        env_thread.start()
        sys_info_thread = threading.Thread(target=report.get_system_info)
        sys_info_thread.start()
    else:
        env_thread, sys_info_thread = None, None

    process.wait()
    report.end_time = time.time()
    lgr.debug("Process ended, setting stop_event to stop monitoring thread")
    stop_event.set()
    if monitoring_thread is not None:
        lgr.debug("Waiting for monitoring thread to finish")
        monitoring_thread.join()
        lgr.debug("Monitoring thread finished")

    # If we have any extra samples that haven't been written yet, do it now
    if report.current_sample is not None:
        report.write_subreport()

    report.process = process
    if env_thread is not None:
        lgr.debug("Waiting for environment collection thread to finish")
        env_thread.join()
        lgr.debug("Environment collection finished")

    if sys_info_thread is not None:
        lgr.debug("Waiting for system information collection thread to finish")
        sys_info_thread.join()
        lgr.debug("System information collection finished")

    if args.record_types.has_system_summary():
        with open(log_paths.info, "a") as system_logs:
            report.run_time_seconds = f"{report.end_time - report.start_time}"
            system_logs.write(report.dump_json())
    safe_close_files(files_to_close)
    lgr.info(report.execution_summary_formatted)
    return report.process.returncode


if __name__ == "__main__":
    main()



from collections import Counter
from copy import deepcopy
from typing import cast
from unittest import mock
import pytest
from con_duct.__main__ import EXECUTION_SUMMARY_FORMAT, ProcessStats, Report, Sample

stat0 = ProcessStats(
    pcpu=0.0,
    pmem=0,
    rss=0,
    vsz=0,
    timestamp="2024-06-11T10:09:37-04:00",
    etime="00:00",
    cmd="cmd 0",
    stat=Counter(["stat0"]),
)

stat1 = ProcessStats(
    pcpu=1.0,
    pmem=1.0,
    rss=1,
    vsz=1,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 1",
    stat=Counter(["stat1"]),
)

stat2 = ProcessStats(
    pcpu=2.0,
    pmem=2.0,
    rss=2,
    vsz=2,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 2",
    stat=Counter(["stat2"]),
)

stat100 = ProcessStats(
    pcpu=100.0,
    pmem=100.0,
    rss=2,
    vsz=2,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 100",
    stat=Counter(["stat100"]),
)
stat_big = ProcessStats(
    pcpu=20000.0,
    pmem=21234234.0,
    rss=43645634562,
    vsz=2345234523452342,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 2",
    stat=Counter(["statbig"]),
)


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_num_samples_increment(mock_log_paths: mock.MagicMock) -> None:
    ex0 = Sample()
    ex0.add_pid(1, deepcopy(stat1))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0
    report.update_from_sample(ex0)
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.current_sample.averages.num_samples == 1
    assert report.full_run_stats.averages.num_samples == 1
    report.update_from_sample(ex0)
    assert report.current_sample.averages.num_samples == 2
    assert report.full_run_stats.averages.num_samples == 2
    report.update_from_sample(ex0)
    assert report.current_sample.averages.num_samples == 3
    assert report.full_run_stats.averages.num_samples == 3


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_single_sample_sanity(mock_log_paths: mock.MagicMock) -> None:
    ex0 = Sample()
    ex0.add_pid(0, deepcopy(stat0))
    ex0.add_pid(1, deepcopy(stat1))
    ex0.add_pid(2, deepcopy(stat2))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0
    report.update_from_sample(ex0)
    # 3 pids in a single sample should still be "1" sample
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.full_run_stats is not None
    assert report.current_sample.averages.num_samples == 1
    assert report.full_run_stats.averages.num_samples == 1

    # assert totals sanity
    assert report.current_sample.total_rss == stat0.rss + stat1.rss + stat2.rss
    assert report.current_sample.total_vsz == stat0.vsz + stat1.vsz + stat2.vsz
    assert report.current_sample.total_pmem == stat0.pmem + stat1.pmem + stat2.pmem
    assert report.current_sample.total_pcpu == stat0.pcpu + stat1.pcpu + stat2.pcpu

    # With one sample averages should be equal to totals
    assert report.current_sample.averages.rss == report.current_sample.averages.rss
    assert report.current_sample.averages.vsz == report.current_sample.averages.vsz
    assert report.current_sample.averages.pmem == report.current_sample.averages.pmem
    assert report.current_sample.averages.pcpu == report.current_sample.averages.pcpu


@pytest.mark.parametrize("stat", [stat0, stat1, stat2, stat_big])
@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_single_stat_multiple_samples_sanity(
    mock_log_paths: mock.MagicMock, stat: ProcessStats
) -> None:
    ex0 = Sample()
    ex0.add_pid(1, deepcopy(stat))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0
    report.update_from_sample(ex0)
    report.update_from_sample(ex0)
    report.update_from_sample(ex0)
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.current_sample.averages.num_samples == 3
    assert report.full_run_stats.averages.num_samples == 3

    # With 3 identical samples, totals should be identical to 1 sample
    assert report.current_sample.total_rss == stat.rss
    assert report.current_sample.total_vsz == stat.vsz
    assert report.current_sample.total_pmem == stat.pmem
    assert report.current_sample.total_pcpu == stat.pcpu

    # Without resetting the current sample, full_run_stats should equal current_sample
    assert report.full_run_stats.total_rss == report.current_sample.total_rss
    assert report.full_run_stats.total_vsz == report.current_sample.total_vsz
    assert report.full_run_stats.total_pmem == report.current_sample.total_pmem
    assert report.full_run_stats.total_pcpu == report.current_sample.total_pcpu
    # Averages too
    assert report.current_sample.averages.rss == report.full_run_stats.averages.rss
    assert report.current_sample.averages.vsz == report.full_run_stats.averages.vsz
    assert report.current_sample.averages.pmem == report.full_run_stats.averages.pmem
    assert report.current_sample.averages.pcpu == report.full_run_stats.averages.pcpu

    # With 3 identical samples, averages should be identical to 1 sample
    assert report.current_sample.averages.rss == report.current_sample.total_rss
    assert report.current_sample.averages.vsz == report.current_sample.total_vsz
    assert report.current_sample.averages.pmem == report.current_sample.total_pmem
    assert report.current_sample.averages.pcpu == report.current_sample.total_pcpu


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_averages(mock_log_paths: mock.MagicMock) -> None:
    sample0 = Sample()
    sample0.add_pid(1, deepcopy(stat0))
    sample1 = Sample()
    sample1.add_pid(1, deepcopy(stat1))
    sample2 = Sample()
    sample2.add_pid(1, deepcopy(stat2))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0
    report.update_from_sample(sample0)
    report.update_from_sample(sample1)
    report.update_from_sample(sample2)
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.current_sample.averages.num_samples == 3
    assert report.full_run_stats.averages.num_samples == 3

    # Assert that average calculation works as expected
    assert (
        report.current_sample.averages.rss == (stat0.rss + stat1.rss + stat2.rss) / 3.0
    )
    assert (
        report.current_sample.averages.vsz == (stat0.vsz + stat1.vsz + stat2.vsz) / 3.0
    )
    assert (
        report.current_sample.averages.pmem
        == (stat0.pmem + stat1.pmem + stat2.pmem) / 3.0
    )
    assert (
        report.current_sample.averages.pcpu
        == (stat0.pcpu + stat1.pcpu + stat2.pcpu) / 3.0
    )
    # And full_run_stats.averages is still identical
    assert report.current_sample.averages.rss == report.full_run_stats.averages.rss
    assert report.current_sample.averages.vsz == report.full_run_stats.averages.vsz
    assert report.current_sample.averages.pmem == report.full_run_stats.averages.pmem
    assert report.current_sample.averages.pcpu == report.full_run_stats.averages.pcpu

    # Lets make the arithmetic a little less round
    report.update_from_sample(sample2)
    report.update_from_sample(sample2)
    report.update_from_sample(sample2)
    assert report.current_sample.averages.num_samples == 6
    assert report.full_run_stats.averages.num_samples == 6
    assert (
        report.current_sample.averages.rss
        == (stat0.rss + stat1.rss + stat2.rss * 4) / 6.0
    )
    assert (
        report.current_sample.averages.vsz
        == (stat0.vsz + stat1.vsz + stat2.vsz * 4) / 6.0
    )
    assert (
        report.current_sample.averages.pmem
        == (stat0.pmem + stat1.pmem + stat2.pmem * 4) / 6.0
    )
    assert (
        report.current_sample.averages.pcpu
        == (stat0.pcpu + stat1.pcpu + stat2.pcpu * 4) / 6.0
    )


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_current_ave_diverges_from_total_ave(
    mock_log_paths: mock.MagicMock,
) -> None:
    sample0 = Sample()
    sample0.add_pid(1, deepcopy(stat0))
    sample1 = Sample()
    sample1.add_pid(1, deepcopy(stat1))
    sample2 = Sample()
    sample2.add_pid(1, deepcopy(stat2))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0
    report.update_from_sample(sample0)
    report.update_from_sample(sample1)
    report.update_from_sample(sample2)
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.current_sample.averages.num_samples == 3
    assert report.full_run_stats.averages.num_samples == 3
    # full_run_stats.averages is still identical to current_sample
    assert report.current_sample.averages.rss == report.full_run_stats.averages.rss
    assert report.current_sample.averages.vsz == report.full_run_stats.averages.vsz
    assert report.current_sample.averages.pmem == report.full_run_stats.averages.pmem
    assert report.current_sample.averages.pcpu == report.full_run_stats.averages.pcpu

    # Reset current_sample so averages will diverge from full_run_stats.averages
    report.current_sample = None
    report.update_from_sample(sample2)
    report.update_from_sample(sample2)
    report.update_from_sample(sample2)
    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    assert report.current_sample.averages.num_samples == 3
    assert report.full_run_stats.averages.num_samples == 6

    # Current sample should only contain sample2
    assert report.current_sample.averages.rss == sample2.total_rss
    assert report.current_sample.averages.vsz == sample2.total_vsz
    assert report.current_sample.averages.pmem == sample2.total_pmem
    assert report.current_sample.averages.pcpu == sample2.total_pcpu

    # Full sample average should == (samples_sum/num_samples)
    assert (
        report.full_run_stats.averages.rss
        == (stat0.rss + stat1.rss + stat2.rss * 4) / 6.0
    )
    assert (
        report.full_run_stats.averages.vsz
        == (stat0.vsz + stat1.vsz + stat2.vsz * 4) / 6.0
    )
    assert (
        report.full_run_stats.averages.pmem
        == (stat0.pmem + stat1.pmem + stat2.pmem * 4) / 6.0
    )
    assert (
        report.full_run_stats.averages.pcpu
        == (stat0.pcpu + stat1.pcpu + stat2.pcpu * 4) / 6.0
    )


@pytest.mark.parametrize("stat", [stat0, stat1, stat2, stat_big])
@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_many_samples(
    mock_log_paths: mock.MagicMock, stat: ProcessStats
) -> None:
    sample1 = Sample()
    pid = 1
    sample1.add_pid(pid, deepcopy(stat))
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    assert report.current_sample is None
    assert report.full_run_stats.averages.num_samples == 0

    # Ensure nothing strange happens after many updates
    for _ in range(100):
        report.update_from_sample(sample1)

    report.current_sample = cast(
        Sample, report.current_sample
    )  # So mypy is convcinced it is not None
    assert report.current_sample is not None
    # Assert that there is exactly 1 ProcessStat.stat count per update
    assert (
        sum(report.current_sample.stats[pid].stat.values())
        == report.full_run_stats.averages.num_samples
        == 100
    )
    assert report.full_run_stats.averages.rss == (stat.rss * 100) / 100.0
    assert report.full_run_stats.averages.vsz == (stat.vsz * 100) / 100.0
    assert report.full_run_stats.averages.pmem == (stat.pmem * 100) / 100.0
    assert report.full_run_stats.averages.pcpu == (stat.pcpu * 100) / 100.0

    # Add a stat that is not 0 and check that the average is still correct
    sample2 = Sample()
    sample2.add_pid(1, deepcopy(stat2))
    report.update_from_sample(sample2)
    assert report.full_run_stats.averages.num_samples == 101
    assert report.full_run_stats.averages.rss == (stat.rss * 100 + stat2.rss) / 101.0
    assert report.full_run_stats.averages.vsz == (stat.vsz * 100 + stat2.vsz) / 101.0
    assert report.full_run_stats.averages.pmem == (stat.pmem * 100 + stat2.pmem) / 101.0
    assert report.full_run_stats.averages.pcpu == (stat.pcpu * 100 + stat2.pcpu) / 101.0


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_sample_no_pids(mock_log_paths: mock.MagicMock) -> None:
    sample0 = Sample()
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    # When there are no pids, finalization should be triggered because the exe is finished,
    # so a Sample with no PIDs should never be passed to update_from_sample.
    with pytest.raises(AssertionError):
        report.update_from_sample(sample0)


@mock.patch("con_duct.__main__.LogPaths")
def test_aggregation_no_false_peak(mock_log_paths: mock.MagicMock) -> None:
    sample1 = Sample()
    sample2 = Sample()
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    sample1.add_pid(1, deepcopy(stat100))
    sample1.add_pid(2, deepcopy(stat0))
    report.update_from_sample(sample1)
    sample2.add_pid(1, deepcopy(stat0))
    sample2.add_pid(2, deepcopy(stat100))
    report.update_from_sample(sample2)
    assert report.current_sample is not None
    assert report.current_sample.total_pcpu == 100



import argparse
from typing import Any
import unittest
from unittest.mock import MagicMock, mock_open, patch
import pytest
from con_duct import suite


class TestSuiteHelpers(unittest.TestCase):

    def test_execute_returns_int(self) -> None:
        def return_non_int(*_args: Any) -> str:
            return "NOPE"

        args = argparse.Namespace(
            command="invalid", file_path="dummy.json", func=return_non_int
        )
        with pytest.raises(TypeError):
            suite.execute(args)


class TestPPrint(unittest.TestCase):

    @patch(
        "builtins.open", new_callable=mock_open, read_data='{"mock_key": "mock_value"}'
    )
    @patch("con_duct.suite.pprint")
    def test_pprint_json(self, mock_pprint: MagicMock, mock_open: MagicMock) -> None:
        args = argparse.Namespace(
            command="pp", file_path="dummy.json", func=suite.pprint_json
        )
        assert suite.execute(args) == 0

        mock_open.assert_called_with("dummy.json", "r")
        mock_pprint.assert_called_once_with({"mock_key": "mock_value"})

    @patch("builtins.open", side_effect=FileNotFoundError)
    def test_file_not_found(self, _mock_open: MagicMock) -> None:
        args = argparse.Namespace(
            command="pp", file_path="dummy.json", func=suite.pprint_json
        )
        assert suite.execute(args) == 1

    @patch("builtins.open", new_callable=mock_open, read_data='{"invalid": "json"')
    @patch("con_duct.suite.pprint")
    def test_pprint_invalid_json(
        self, mock_pprint: MagicMock, mock_open: MagicMock
    ) -> None:
        args = argparse.Namespace(
            command="pp", file_path="dummy.json", func=suite.pprint_json
        )
        assert suite.execute(args) == 1

        mock_open.assert_called_with("dummy.json", "r")
        mock_pprint.assert_not_called()



from __future__ import annotations
from dataclasses import asdict
import os
import re
from unittest.mock import MagicMock, call, patch
import pytest
from con_duct.__main__ import LogPaths, Outputs


def test_log_paths_filesafe_datetime_prefix() -> None:
    log_paths = LogPaths.create("start_{datetime_filesafe}")
    pattern = r"^start_\d{4}\.\d{2}\.\d{2}T\d{2}\.\d{2}\.\d{2}.*"
    for path in asdict(log_paths).values():
        assert re.match(pattern, path) is not None


def test_log_paths_pid_prefix() -> None:
    prefix = "prefix_{pid}_"
    log_paths = LogPaths.create(prefix, pid=123456)
    assert log_paths.prefix == prefix.format(pid=123456)


@pytest.mark.parametrize(
    "path",
    [
        "directory/",
        "nested/directory/",
        "/abs/path/",
    ],
)
@patch("con_duct.__main__.os.makedirs")
@patch("con_duct.__main__.os.path.exists")
@patch("builtins.open")
def test_prepare_dir_paths_available(
    _mock_open: MagicMock, mock_exists: MagicMock, mock_mkdir: MagicMock, path: str
) -> None:
    mock_exists.return_value = False
    log_paths = LogPaths.create(path)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.ALL)
    mock_mkdir.assert_called_once_with(path, exist_ok=True)


@pytest.mark.parametrize(
    "path",
    [
        "directory/pre_",
        "nested/directory/pre_",
        "/abs/path/pre_",
    ],
)
@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prefix_with_filepart_and_directory_part(
    mock_open: MagicMock, mock_mkdir: MagicMock, mock_exists: MagicMock, path: str
) -> None:
    mock_exists.return_value = False
    log_paths = LogPaths.create(path)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.ALL)
    mock_mkdir.assert_called_once_with(os.path.dirname(path), exist_ok=True)
    expected_calls = [call(each, "w") for _name, each in log_paths]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prefix_with_filepart_only(
    mock_open: MagicMock, mock_mkdir: MagicMock, mock_exists: MagicMock
) -> None:
    mock_exists.return_value = False
    log_paths = LogPaths.create("filepartonly")
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.ALL)
    mock_mkdir.assert_not_called()
    expected_calls = [call(each, "w") for _name, each in log_paths]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prepare_file_paths_available_all(
    mock_open: MagicMock, _mock_mkdir: MagicMock, mock_exists: MagicMock
) -> None:
    mock_exists.return_value = False
    prefix = "prefix_"
    log_paths = LogPaths.create(prefix)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.ALL)
    expected_calls = [call(each, "w") for _name, each in log_paths]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prepare_file_paths_available_stdout(
    mock_open: MagicMock, _mock_mkdir: MagicMock, mock_exists: MagicMock
) -> None:
    mock_exists.return_value = False
    prefix = "prefix_"
    log_paths = LogPaths.create(prefix)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.STDOUT)
    expected_calls = [
        call(each, "w") for name, each in log_paths if name != Outputs.STDERR
    ]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prepare_file_paths_available_stderr(
    mock_open: MagicMock, _mock_mkdir: MagicMock, mock_exists: MagicMock
) -> None:
    mock_exists.return_value = False
    prefix = "prefix_"
    log_paths = LogPaths.create(prefix)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.STDERR)
    expected_calls = [
        call(each, "w") for name, each in log_paths if name != Outputs.STDOUT
    ]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.path.exists")
@patch("con_duct.__main__.os.makedirs")
@patch("builtins.open")
def test_prepare_file_paths_available_no_streams(
    mock_open: MagicMock, _mock_mkdir: MagicMock, mock_exists: MagicMock
) -> None:
    mock_exists.return_value = False
    prefix = "prefix_"
    log_paths = LogPaths.create(prefix)
    log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.NONE)
    streams = [Outputs.STDOUT, Outputs.STDERR]
    expected_calls = [
        call(each, "w") for name, each in log_paths if name not in streams
    ]
    mock_open.assert_has_calls(expected_calls, any_order=True)


@patch("con_duct.__main__.os.makedirs")
@patch("con_duct.__main__.os.path.exists")
@patch("builtins.open")
def test_prepare_paths_not_available_no_clobber(
    mock_open: MagicMock, mock_exists: MagicMock, mock_mkdir: MagicMock
) -> None:
    mock_exists.return_value = True
    log_paths = LogPaths.create("doesntmatter")
    with pytest.raises(FileExistsError):
        log_paths.prepare_paths(clobber=False, capture_outputs=Outputs.ALL)
    mock_mkdir.assert_not_called()
    mock_open.assert_not_called()



from __future__ import annotations
from collections import Counter
from copy import deepcopy
from datetime import datetime
import os
import subprocess
from unittest import mock
import pytest
from con_duct.__main__ import (
    EXECUTION_SUMMARY_FORMAT,
    Averages,
    ProcessStats,
    Report,
    Sample,
)

stat0 = ProcessStats(
    pcpu=0.0,
    pmem=0,
    rss=0,
    vsz=0,
    timestamp="2024-06-11T10:09:37-04:00",
    etime="00:00",
    cmd="cmd 1",
    stat=Counter(["stat0"]),
)

stat1 = ProcessStats(
    pcpu=1.0,
    pmem=0,
    rss=0,
    vsz=0,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 1",
    stat=Counter(["stat1"]),
)

stat2 = ProcessStats(
    pcpu=1.1,
    pmem=1.1,
    rss=11,
    vsz=11,
    timestamp="2024-06-11T10:13:23-04:00",
    etime="00:02",
    cmd="cmd 1",
    stat=Counter(["stat2"]),
)


def test_sample_max_initial_values_one_pid() -> None:
    maxes = Sample()
    ex0 = Sample()
    ex0.add_pid(1, deepcopy(stat0))
    maxes = maxes.aggregate(ex0)
    assert maxes.stats == {1: stat0}


def test_sample_max_one_pid() -> None:
    maxes = Sample()
    maxes.add_pid(1, deepcopy(stat0))
    ex1 = Sample()
    ex1.add_pid(1, deepcopy(stat1))
    maxes = maxes.aggregate(ex1)
    assert maxes.stats[1].rss == stat1.rss
    assert maxes.stats[1].vsz == stat1.vsz
    assert maxes.stats[1].pmem == stat1.pmem
    assert maxes.stats[1].pcpu == stat1.pcpu


def test_sample_max_initial_values_two_pids() -> None:
    maxes = Sample()
    ex0 = Sample()
    ex0.add_pid(1, deepcopy(stat0))
    ex0.add_pid(2, deepcopy(stat0))
    maxes = maxes.aggregate(ex0)
    assert maxes.stats == {1: stat0, 2: stat0}
    assert maxes.stats == {1: stat0, 2: stat0}


def test_sample_aggregate_two_pids() -> None:
    maxes = Sample()
    maxes.add_pid(1, deepcopy(stat0))
    maxes.add_pid(2, deepcopy(stat0))
    assert maxes.stats[1].stat["stat0"] == 1
    assert maxes.stats[2].stat["stat0"] == 1
    assert maxes.stats[1].stat["stat1"] == 0
    assert maxes.stats[2].stat["stat1"] == 0
    ex1 = Sample()
    ex1.add_pid(1, deepcopy(stat1))
    maxes = maxes.aggregate(ex1)
    assert maxes.stats[1].stat["stat0"] == 1
    assert maxes.stats[2].stat["stat0"] == 1
    assert maxes.stats[1].stat["stat1"] == 1
    assert maxes.stats[2].stat["stat1"] == 0
    ex2 = Sample()
    ex2.add_pid(2, deepcopy(stat1))
    maxes = maxes.aggregate(ex2)
    # Check the `stat` counts one of each for both pids
    assert maxes.stats[1].stat["stat0"] == 1
    assert maxes.stats[2].stat["stat0"] == 1
    assert maxes.stats[1].stat["stat1"] == 1
    assert maxes.stats[2].stat["stat1"] == 1

    # Each stat1 value > stat0 value
    assert maxes.stats[1].pcpu == stat1.pcpu
    assert maxes.stats[1].pmem == stat1.pmem
    assert maxes.stats[1].rss == stat1.rss
    assert maxes.stats[1].vsz == stat1.vsz
    assert maxes.stats[2].pcpu == stat1.pcpu
    assert maxes.stats[2].pmem == stat1.pmem
    assert maxes.stats[2].rss == stat1.rss
    assert maxes.stats[2].vsz == stat1.vsz


def test_average_no_samples() -> None:
    averages = Averages()
    assert averages.num_samples == 0
    sample = Sample()
    sample.averages = averages
    serialized = sample.for_json()
    assert "averages" in serialized
    assert not serialized["averages"]


def test_averages_one_sample() -> None:
    sample = Sample()
    sample.add_pid(1, deepcopy(stat0))
    averages = Averages.from_sample(sample)
    assert averages.rss == sample.total_rss
    assert averages.vsz == sample.total_vsz
    assert averages.pmem == sample.total_pmem
    assert averages.pcpu == sample.total_pcpu
    assert averages.num_samples == 1


def test_averages_two_samples() -> None:
    sample = Sample()
    sample.add_pid(1, deepcopy(stat0))
    averages = Averages.from_sample(sample)
    sample2 = Sample()
    sample2.add_pid(2, deepcopy(stat1))
    averages.update(sample2)
    assert averages.pcpu == (stat0.pcpu + stat1.pcpu) / 2


def test_averages_three_samples() -> None:
    sample = Sample()
    sample.add_pid(1, deepcopy(stat0))
    averages = Averages.from_sample(sample)
    sample2 = Sample()
    sample2.add_pid(2, deepcopy(stat1))
    averages.update(sample2)
    averages.update(sample2)
    assert averages.pcpu == (stat0.pcpu + (2 * stat1.pcpu)) / 3


def test_sample_totals() -> None:
    sample = Sample()
    sample.add_pid(1, deepcopy(stat2))
    sample.add_pid(2, deepcopy(stat2))
    assert sample.total_rss == stat2.rss * 2
    assert sample.total_vsz == stat2.vsz * 2
    assert sample.total_pmem == stat2.pmem * 2
    assert sample.total_pcpu == stat2.pcpu * 2


@pytest.mark.parametrize(
    "pcpu, pmem, rss, vsz, etime, cmd",
    [
        (1.0, 1.1, 1024, 1025, "00:00", "cmd"),
        (0.5, 0.7, 20.48, 40.96, "00:01", "any"),
        (1, 2, 3, 4, "100:1000", "string"),
        (0, 0.0, 0, 0.0, "999:999:999", "can have spaces"),
        (2.5, 3.5, 8192, 16384, "any", "for --this --kind of thing"),
        (100.0, 99.9, 65536, 131072, "string", "cmd"),
    ],
)
def test_process_stats_green(
    pcpu: float, pmem: float, rss: int, vsz: int, etime: str, cmd: str
) -> None:
    # Assert does not raise
    ProcessStats(
        pcpu=pcpu,
        pmem=pmem,
        rss=rss,
        vsz=vsz,
        timestamp=datetime.now().astimezone().isoformat(),
        etime=etime,
        cmd=cmd,
        stat=Counter(["stat0"]),
    )


@pytest.mark.parametrize(
    "pcpu, pmem, rss, vsz, etime, cmd",
    [
        ("only", 1.1, 1024, 1025, "etime", "cmd"),
        (0.5, "takes", 20.48, 40.96, "some", "str"),
        (1, 2, "one", 4, "anything", "accepted"),
        (1, 2, 3, "value", "etime", "cmd"),
        ("2", "fail", "or", "more", "etime", "cmd"),
    ],
)
def test_process_stats_red(
    pcpu: float, pmem: float, rss: int, vsz: int, etime: str, cmd: str
) -> None:
    with pytest.raises(AssertionError):
        ProcessStats(
            pcpu=pcpu,
            pmem=pmem,
            rss=rss,
            vsz=vsz,
            timestamp=datetime.now().astimezone().isoformat(),
            etime=etime,
            cmd=cmd,
            stat=Counter(["stat0"]),
        )


@mock.patch("con_duct.__main__.LogPaths")
def test_system_info_sanity(mock_log_paths: mock.MagicMock) -> None:
    mock_log_paths.prefix = "mock_prefix"
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    report.get_system_info()
    assert report.system_info is not None
    assert report.system_info.hostname
    assert report.system_info.cpu_total
    assert report.system_info.memory_total > 10
    assert report.system_info.uid == os.getuid()
    assert report.system_info.user == os.environ.get("USER")


@mock.patch("con_duct.__main__.shutil.which")
@mock.patch("con_duct.__main__.subprocess.check_output")
@mock.patch("con_duct.__main__.LogPaths")
def test_gpu_parsing_green(
    mock_log_paths: mock.MagicMock, mock_sp: mock.MagicMock, _mock_which: mock.MagicMock
) -> None:
    mock_sp.return_value = (
        "index, name, pci.bus_id, driver_version, memory.total [MiB], compute_mode\n"
        "0, NVIDIA RTX A5500 Laptop GPU, 00000000:01:00.0, 535.183.01, 16384 MiB, Default"
    ).encode("utf-8")
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    report.get_system_info()
    assert report.gpus is not None
    assert report.gpus == [
        {
            "index": "0",
            "name": "NVIDIA RTX A5500 Laptop GPU",
            "bus_id": "00000000:01:00.0",
            "driver_version": "535.183.01",
            "memory.total": "16384 MiB",
            "compute_mode": "Default",
        }
    ]


@mock.patch("con_duct.__main__.lgr")
@mock.patch("con_duct.__main__.shutil.which")
@mock.patch("con_duct.__main__.subprocess.check_output")
@mock.patch("con_duct.__main__.LogPaths")
def test_gpu_call_error(
    mock_log_paths: mock.MagicMock,
    mock_sp: mock.MagicMock,
    _mock_which: mock.MagicMock,
    mlgr: mock.MagicMock,
) -> None:
    mock_sp.side_effect = subprocess.CalledProcessError(1, "errrr")
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    report.get_system_info()
    assert report.gpus is None
    mlgr.warning.assert_called_once()


@mock.patch("con_duct.__main__.lgr")
@mock.patch("con_duct.__main__.shutil.which")
@mock.patch("con_duct.__main__.subprocess.check_output")
@mock.patch("con_duct.__main__.LogPaths")
def test_gpu_parse_error(
    mock_log_paths: mock.MagicMock,
    mock_sp: mock.MagicMock,
    _mock_which: mock.MagicMock,
    mlgr: mock.MagicMock,
) -> None:
    mock_sp.return_value = (
        "index, name, pci.bus_id, driver_version, memory.total [MiB], compute_mode\n"
        "not-enough-values, 535.183.01, 16384 MiB, Default"
    ).encode("utf-8")
    report = Report("_cmd", [], mock_log_paths, EXECUTION_SUMMARY_FORMAT, clobber=False)
    report.get_system_info()
    assert report.gpus is None
    mlgr.warning.assert_called_once()



import argparse
import pytest
from con_duct.__main__ import Arguments, assert_num


def test_sample_less_than_report_interval() -> None:
    args = Arguments.from_argv(
        ["fake"],
        sample_interval=0.01,
        report_interval=0.1,
    )
    assert args.sample_interval <= args.report_interval


def test_sample_equal_to_report_interval() -> None:
    args = Arguments.from_argv(
        ["fake"],
        sample_interval=0.1,
        report_interval=0.1,
    )
    assert args.sample_interval == args.report_interval


def test_sample_equal_greater_than_report_interval() -> None:
    with pytest.raises(argparse.ArgumentError):
        Arguments.from_argv(
            ["fake"],
            sample_interval=1.0,
            report_interval=0.1,
        )


@pytest.mark.parametrize("input_value", [0, 1, 2, -1, 100, 0.001, -1.68])
def test_assert_num_green(input_value: int) -> None:
    assert_num(input_value)


@pytest.mark.parametrize("input_value", ["hi", "0", "one"])
def test_assert_num_red(input_value: int) -> None:
    with pytest.raises(AssertionError):
        assert_num(input_value)



from __future__ import annotations
import subprocess
from unittest.mock import MagicMock, call, patch
from utils import MockStream
from con_duct.__main__ import LogPaths, Outputs, prepare_outputs


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_none_output_none(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.NONE, Outputs.NONE, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_not_called()
    assert stdout == subprocess.DEVNULL
    assert stderr == subprocess.DEVNULL


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_none_output_stdout(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.NONE, Outputs.STDOUT, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_not_called()
    assert stdout is None
    assert stderr == subprocess.DEVNULL


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_none_output_stderr(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.NONE, Outputs.STDERR, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_not_called()
    assert stdout == subprocess.DEVNULL
    assert stderr is None


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_none_output_all(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.NONE, Outputs.ALL, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_not_called()
    assert stdout is None
    assert stderr is None


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stdout_output_none(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDOUT, Outputs.NONE, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_called_once_with(mock_log_paths.stdout, "w")
    assert stdout == mock_open.return_value
    assert stderr == subprocess.DEVNULL


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stdout_output_stdout(
    mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDOUT, Outputs.STDOUT, mock_log_paths)
    mock_tee_stream.assert_called_once_with(
        mock_log_paths.stdout, buffer=mock_stdout.buffer
    )
    mock_open.assert_not_called()
    assert stdout == mock_tee_stream.return_value
    assert stderr == subprocess.DEVNULL


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stdout_output_stderr(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDOUT, Outputs.STDERR, mock_log_paths)
    mock_open.assert_called_once_with(mock_log_paths.stdout, "w")
    mock_tee_stream.assert_not_called()
    assert stdout == mock_open.return_value
    assert stderr is None


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stdout_output_all(
    mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDOUT, Outputs.ALL, mock_log_paths)
    mock_tee_stream.assert_called_once_with(
        mock_log_paths.stdout, buffer=mock_stdout.buffer
    )
    mock_open.assert_not_called()
    assert stdout == mock_tee_stream.return_value
    assert stderr is None


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stderr_output_none(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDERR, Outputs.NONE, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_called_once_with(mock_log_paths.stderr, "w")
    assert stdout == subprocess.DEVNULL
    assert stderr == mock_open.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stderr_output_stdout(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDERR, Outputs.STDOUT, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_called_once_with(mock_log_paths.stderr, "w")
    assert stdout is None
    assert stderr == mock_open.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stderr_output_stderr(
    _mock_stdout: MockStream,
    mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDERR, Outputs.STDERR, mock_log_paths)
    mock_tee_stream.assert_called_once_with(
        mock_log_paths.stderr, buffer=mock_stderr.buffer
    )
    mock_open.assert_not_called()
    assert stdout == subprocess.DEVNULL
    assert stderr == mock_tee_stream.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_stderr_output_all(
    _mock_stdout: MockStream,
    mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.STDERR, Outputs.ALL, mock_log_paths)
    mock_tee_stream.assert_called_once_with(
        mock_log_paths.stderr, buffer=mock_stderr.buffer
    )
    mock_open.assert_not_called()
    assert stdout is None
    assert stderr == mock_tee_stream.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_all_output_none(
    _mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.ALL, Outputs.NONE, mock_log_paths)
    mock_tee_stream.assert_not_called()
    mock_open.assert_has_calls(
        [call(mock_log_paths.stdout, "w"), call(mock_log_paths.stderr, "w")]
    )
    assert stdout == mock_open.return_value
    assert stderr == mock_open.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_all_output_stdout(
    mock_stdout: MockStream,
    _mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.ALL, Outputs.STDOUT, mock_log_paths)
    mock_tee_stream.assert_called_with(mock_log_paths.stdout, buffer=mock_stdout.buffer)
    mock_open.assert_called_once_with(mock_log_paths.stderr, "w")
    assert stdout == mock_tee_stream.return_value
    assert stderr == mock_open.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_all_output_stderr(
    _mock_stdout: MockStream,
    mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.ALL, Outputs.STDERR, mock_log_paths)
    mock_tee_stream.assert_called_once_with(
        mock_log_paths.stderr, buffer=mock_stderr.buffer
    )
    mock_open.assert_called_once_with(mock_log_paths.stdout, "w")
    assert stdout == mock_open.return_value
    assert stderr == mock_tee_stream.return_value


@patch("builtins.open", new_callable=MagicMock)
@patch("con_duct.__main__.TailPipe")
@patch("con_duct.__main__.LogPaths")
@patch("con_duct.__main__.sys.stderr", new_callable=MockStream)
@patch("con_duct.__main__.sys.stdout", new_callable=MockStream)
def test_prepare_outputs_capture_all_output_all(
    mock_stdout: MockStream,
    mock_stderr: MockStream,
    mock_LogPaths: LogPaths,
    mock_tee_stream: MagicMock,
    mock_open: MagicMock,
) -> None:
    mock_log_paths = mock_LogPaths.create("mock_prefix")
    mock_tee_stream.return_value.start = MagicMock()
    stdout, stderr = prepare_outputs(Outputs.ALL, Outputs.ALL, mock_log_paths)
    mock_tee_stream.assert_has_calls(
        [
            call(mock_log_paths.stdout, buffer=mock_stdout.buffer),
            call(mock_log_paths.stderr, buffer=mock_stderr.buffer),
        ],
        any_order=True,
    )
    assert stdout == mock_tee_stream.return_value
    assert stderr == mock_tee_stream.return_value
    mock_open.assert_not_called()
